@@ -16,7 +16,7 @@
 DSET extended from DSNT (soft-argmax) operations for use in PyTorch computation graphs.
 """
 
-__version__ = "0.1.3"
+__version__ = "0.1.5"
 __author__ = 'Albert Garcia'
 __credits__ = 'SnT - Interdisciplinary Centre for Security, Reliability and Trust'
 
@@ -897,3 +897,298 @@ def batched_gaussian_symmetric_lines_completion(heatmaps, padding_target=None, s
     
     return torch.stack(all_heatmaps, dim=0)
 
+#------------- SUPPORT FUNCTIONS FOR FAST GAUSSIAN HEATMAP COMPLETION (symmetric image implementation) ------------------#
+
+def maximum_along_border(heatmap, not_filled_values, dilation_kernel):
+    not_filled_values_dilated = not_filled_values.clone().unsqueeze(0).unsqueeze(0).float()
+    not_filled_values_dilated = torch.nn.functional.conv2d(not_filled_values_dilated, dilation_kernel, padding=1)[0, 0] >= 1
+    return heatmap[not_filled_values_dilated].max()
+
+#------------------------------------------------------------------------------------------------------------------------#
+
+def gaussian_symmetric_image(generated_heatmap, padding_target=None, dilation_iters=0, alpha=0.25, return_filled=False, make_plots=False, overwrite_original_pixels=False, normalize=True):
+    """Perform fast Gaussian completion on a heatmap based on the symmetric lines that cross the center of the Gaussian distribution (maximum pixel). This implementation makes use of coordinates masks for fast completion with an optional dilation completion process.
+
+    Args:
+        generated_heatmap (torch.Tensor): Heatmap containing a Gaussian distribution to be completed/extended with shape [H, W]
+        padding_target (tuple, default = None): Padding to be added and filled to the input heatmap in the form of the tuple (left, right, top, bottom). If None, then the tuple defaults to (W, W, H, H). Maximum padding is (W, W, H, H).
+        dilation_iters (int, default = 0): Number of times to repeat the process of heatmap dilation used to further extend the symmetrized heatmap.
+        alpha (float, default = 0.25): Additional value to be added to the denumerator during the averaging of the pixels within the pixel dilation process.
+        return_filled (bool, default = True): Whether to return a boolean mask with all the pixels that have been filled over the extended heatmap.
+        make_plots (bool, default = True): Whether to show plots of the diverse heatmaps through the process.
+        overwrite_original_pixels (bool, default = False): Whether to overwrite the pixels of the input heatmap. This can affect ellipse perfomance due to modifying the original heatmap.
+        normalize (bool, default = True): Whether to normalize the extended heatmap to be returned.
+
+    Returns:
+        extended_gaussian_heatmap (torch.Tensor): Extended heatmap containing a completed Gaussian distribution with shape [top + H + bottom, left + W + right]
+        filled_values (torch.Tensor): Optional boolean mask with all the pixels that have been filled over the extended heatmap with shape [top + H + bottom, left + W + right]
+    """
+    assert isinstance(dilation_iters, int),'Dilation iters must be an integer'
+    assert dilation_iters>=0,'Dilation iters must be an positive'
+    
+    device = generated_heatmap.device
+    h, w = generated_heatmap.shape[-2:]
+
+    default_padding = (w,w,h,h)
+    extended_gaussian_heatmap = torch.nn.functional.pad(generated_heatmap, default_padding).to(device)
+    ext_h, ext_w = extended_gaussian_heatmap.shape[-2:]
+    original_center = default_padding[2], default_padding[0]
+
+    gaussian_center = generated_heatmap.argmax()
+    gaussian_center = torch.Tensor( [gaussian_center//w, gaussian_center%w] ).to(device)
+
+    x_mask = torch.arange(0, w).to(device)
+    y_mask = torch.arange(0, h).to(device)
+    
+    x_mask = x_mask + 2*(gaussian_center[1] - x_mask) + w
+    y_mask = y_mask + 2*(gaussian_center[0] - y_mask) + h
+    
+    x_mask = x_mask.unsqueeze(0).repeat(h, 1).long()
+    y_mask = y_mask.unsqueeze(1).repeat(1, w).long()
+    
+    if overwrite_original_pixels:
+        extended_gaussian_heatmap[y_mask, x_mask] = generated_heatmap
+    else:
+        not_filled_values = (extended_gaussian_heatmap<=0).to(device)
+        temporal_extended_heatmap = torch.zeros(extended_gaussian_heatmap.shape).to(device)
+        temporal_extended_heatmap[y_mask, x_mask] = generated_heatmap
+        extended_gaussian_heatmap[not_filled_values] = temporal_extended_heatmap[not_filled_values]
+    
+    not_filled_values = (extended_gaussian_heatmap<=0).to(device)
+    
+    if make_plots:
+        plt.imshow(extended_gaussian_heatmap.detach().cpu())
+        plt.title('Extended heatmap before dilation')
+        plt.show()
+    
+    if dilation_iters>0:
+        dilation_kernel = torch.ones((1,1,3,3)).to(device)
+        dilated_gaussian_heatmap = extended_gaussian_heatmap.clone().to(device).unsqueeze(0).unsqueeze(0)
+        dilated_filled_values = (extended_gaussian_heatmap>0).to(device).unsqueeze(0).unsqueeze(0).float()
+        for n in range(dilation_iters):
+            dilated_gaussian_heatmap = torch.nn.functional.conv2d(dilated_gaussian_heatmap, dilation_kernel, padding=1)
+            dilated_filled_values = torch.nn.functional.conv2d(dilated_filled_values, dilation_kernel, padding=1)
+            
+            dilated_filled_values = dilated_filled_values + alpha
+            dilated_gaussian_heatmap = dilated_gaussian_heatmap / dilated_filled_values
+            dilated_filled_values = (dilated_gaussian_heatmap>0).float()
+            
+
+        dilated_gaussian_heatmap = dilated_gaussian_heatmap[0,0]
+
+        if make_plots:
+            plt.title('Dilated heatmap')
+            plt.imshow(dilated_gaussian_heatmap.detach().cpu())
+            plt.show()
+    
+        max_heatmap_border = maximum_along_border(extended_gaussian_heatmap, not_filled_values, dilation_kernel).to(device)
+        max_dilated = dilated_gaussian_heatmap[not_filled_values].max().to(device)
+        dilated_gaussian_heatmap = dilated_gaussian_heatmap / max_dilated
+        dilated_gaussian_heatmap = dilated_gaussian_heatmap * max_heatmap_border
+        
+        dilated_gaussian_heatmap[~not_filled_values] = 0
+        extended_gaussian_heatmap = extended_gaussian_heatmap + dilated_gaussian_heatmap
+        
+    if padding_target is not None:
+        left, right, top, bottom = padding_target
+        assert left<=w and right<=w and top<=h and bottom<=h,'Maximum padding allowed is (W,W,H,H)'
+        extended_gaussian_heatmap = extended_gaussian_heatmap[h-top:2*h+bottom, w-left:2*w+right]
+    
+    if normalize: extended_gaussian_heatmap = extended_gaussian_heatmap/extended_gaussian_heatmap.sum()
+    
+    filled_values = (extended_gaussian_heatmap>0).to(device)
+    
+    if make_plots:
+        fig, ax = plt.subplots(1, 3, figsize=(15, 15))
+        ax[0].imshow(generated_heatmap.detach().cpu())
+        ax[0].set_title('Original heatmap')
+        ax[1].imshow(filled_values.detach().cpu())
+        ax[1].set_title('Filled pixels')
+        ax[2].imshow(extended_gaussian_heatmap.detach().cpu())
+        ax[2].set_title('Final extended heatmap')
+        plt.show()
+    
+    if return_filled: return extended_gaussian_heatmap, filled_values
+    return extended_gaussian_heatmap
+
+def batched_gaussian_symmetric_image(heatmaps, padding_target=None, dilation_iters=0, alpha=0.25, overwrite_original_pixels=False, normalize=True):
+    """Perform fast Gaussian completion on a batch of heatmaps based on the symmetric lines that cross the center of the Gaussian distribution (maximum pixel). This implementation makes use of coordinates masks for fast completion with an optional dilation completion process.
+
+    Args:
+        heatmaps (torch.Tensor): Heatmaps generated by the model with shape [B, E, H, W]
+        padding_target (tuple, default = None): Padding to be added and filled to the input heatmap in the form of the tuple (left, right, top, bottom). If None, then the tuple defaults to (W, W, H, H). Maximum padding is (W, W, H, H).
+        dilation_iters (int, default = 0): Number of times to repeat the process of heatmap dilation used to further extend the symmetrized heatmap.
+        alpha (float, default = 0.25): Additional value to be added to the denumerator during the averaging of the pixels within the pixel dilation process.
+        overwrite_original_pixels (bool, default = False): Whether to overwrite the pixels of the input heatmap. This can affect ellipse perfomance due to modifying the original heatmap.
+        normalize (bool, default = True): Whether to normalize the extended heatmap to be returned.
+
+    Returns:
+        extended_gaussian_heatmap (torch.Tensor): Extended heatmap containing a completed Gaussian distribution with shape [top + H + bottom, left + W + right]
+        filled_values (torch.Tensor): Optional boolean mask with all the pixels that have been filled over the extended heatmap with shape [top + H + bottom, left + W + right]
+    """
+    
+    batches, locations = heatmaps.size()[0:2]
+    all_heatmaps = []
+    
+    for batch_num in range(batches):
+        batch_heatmaps = []
+        
+        for location_num in range(locations):
+            extended_heatmap = gaussian_symmetric_image(heatmaps[batch_num, location_num], padding_target=padding_target, dilation_iters=dilation_iters, alpha=alpha, overwrite_original_pixels=overwrite_original_pixels, normalize=normalize)
+            
+            batch_heatmaps.append( extended_heatmap )
+        
+        batch_heatmaps =  torch.stack(batch_heatmaps, dim=0)
+        all_heatmaps.append( batch_heatmaps )
+    
+    return torch.stack(all_heatmaps, dim=0)
+
+#---------------------------------------------------------------------------------------------------------------------------------------------------------------------#
+
+def gaussian_symmetric_wave_image(generated_heatmap, padding_target=None, dilation_iters=0, alpha=0.25, return_filled=False, make_plots=False, overwrite_original_pixels=False, normalize=True):
+    """Perform fast Gaussian completion on a heatmap based on the symmetric lines that cross the center of the Gaussian distribution (maximum pixel). This implementation makes use of coordinates masks for fast completion with an optional dilation completion process. The dilation completion process makes use of a wave mask that dilatates at every iteration ensuring a continuous extension of the heatmap.
+
+    Args:
+        generated_heatmap (torch.Tensor): Heatmap containing a Gaussian distribution to be completed/extended with shape [H, W]
+        padding_target (tuple, default = None): Padding to be added and filled to the input heatmap in the form of the tuple (left, right, top, bottom). If None, then the tuple defaults to (W, W, H, H). Maximum padding is (W, W, H, H).
+        dilation_iters (int, default = 0): Number of times to repeat the process of heatmap dilation used to further extend the symmetrized heatmap.
+        alpha (float, default = 0.25): Additional value to be added to the denumerator during the averaging of the pixels within the pixel dilation process.
+        return_filled (bool, default = True): Whether to return a boolean mask with all the pixels that have been filled over the extended heatmap.
+        make_plots (bool, default = True): Whether to show plots of the diverse heatmaps through the process.
+        overwrite_original_pixels (bool, default = False): Whether to overwrite the pixels of the input heatmap. This can affect ellipse perfomance due to modifying the original heatmap.
+        normalize (bool, default = True): Whether to normalize the extended heatmap to be returned.
+
+    Returns:
+        extended_gaussian_heatmap (torch.Tensor): Extended heatmap containing a completed Gaussian distribution with shape [top + H + bottom, left + W + right]
+        filled_values (torch.Tensor): Optional boolean mask with all the pixels that have been filled over the extended heatmap with shape [top + H + bottom, left + W + right]
+    """
+    assert isinstance(dilation_iters, int),'Dilation iters must be an integer'
+    assert dilation_iters>=0,'Dilation iters must be an positive'
+    assert isinstance(alpha, float),'Alpha must be a float'
+    assert alpha>0,'Alpha must be strictly positive'
+    
+    device = generated_heatmap.device
+    h, w = generated_heatmap.shape[-2:]
+
+    default_padding = (w,w,h,h)
+    extended_gaussian_heatmap = torch.nn.functional.pad(generated_heatmap, default_padding).to(device)
+    ext_h, ext_w = extended_gaussian_heatmap.shape[-2:]
+    original_center = default_padding[2], default_padding[0]
+
+    gaussian_center = generated_heatmap.argmax()
+    gaussian_center = torch.Tensor( [gaussian_center//w, gaussian_center%w] ).to(device)
+
+    x_mask = torch.arange(0, w).to(device)
+    y_mask = torch.arange(0, h).to(device)
+    
+    x_mask = x_mask + 2*(gaussian_center[1] - x_mask) + w
+    y_mask = y_mask + 2*(gaussian_center[0] - y_mask) + h
+    
+    x_mask = x_mask.unsqueeze(0).repeat(h, 1).long()
+    y_mask = y_mask.unsqueeze(1).repeat(1, w).long()
+    
+    if overwrite_original_pixels:
+        extended_gaussian_heatmap[y_mask, x_mask] = generated_heatmap
+    else:
+        not_filled_values = (extended_gaussian_heatmap<=0).to(device)
+        temporal_extended_heatmap = torch.zeros(extended_gaussian_heatmap.shape).to(device)
+        temporal_extended_heatmap[y_mask, x_mask] = generated_heatmap
+        extended_gaussian_heatmap[not_filled_values] = temporal_extended_heatmap[not_filled_values]
+    
+    not_filled_values = (extended_gaussian_heatmap<=0).to(device)
+    
+    if make_plots:
+        plt.imshow(extended_gaussian_heatmap.detach().cpu())
+        plt.title('Extended heatmap before dilation')
+        plt.show()
+    
+    if dilation_iters>0:
+        dilation_kernel = torch.ones((1,1,3,3)).to(device)
+        dilated_gaussian_heatmap = extended_gaussian_heatmap.clone().to(device).unsqueeze(0).unsqueeze(0)
+        wave_gaussian_heatmap = torch.zeros(dilated_gaussian_heatmap.shape).to(device)
+        boolean_dilated_values = (extended_gaussian_heatmap>0).to(device).unsqueeze(0).unsqueeze(0).float()
+        dilated_filled_values = boolean_dilated_values.clone().to(device)
+        
+        for n in range(dilation_iters):
+            dilated_gaussian_heatmap = torch.nn.functional.conv2d(dilated_gaussian_heatmap, dilation_kernel, padding=1)
+            dilated_filled_values = torch.nn.functional.conv2d(dilated_filled_values, dilation_kernel, padding=1)
+            new_boolean_dilated_values = torch.clamp( torch.nn.functional.conv2d(boolean_dilated_values, dilation_kernel, padding=1), 0, 1)
+            wave_idxs = ((new_boolean_dilated_values - boolean_dilated_values)==1).to(device)
+            
+            boolean_dilated_values = new_boolean_dilated_values
+            dilated_filled_values = dilated_filled_values + alpha
+            dilated_gaussian_heatmap = dilated_gaussian_heatmap / dilated_filled_values
+            dilated_filled_values = (dilated_gaussian_heatmap>0).to(device).float()
+            
+            wave_gaussian_heatmap[wave_idxs] = dilated_gaussian_heatmap[wave_idxs]
+
+
+        dilated_gaussian_heatmap = dilated_gaussian_heatmap[0,0]
+        wave_gaussian_heatmap = wave_gaussian_heatmap[0,0]
+
+        if make_plots:
+            plt.title('Dilated heatmap')
+            plt.imshow(dilated_gaussian_heatmap.detach().cpu())
+            plt.show()
+            
+            plt.title('Wave dilated heatmap')
+            plt.imshow(wave_gaussian_heatmap.detach().cpu())
+            plt.show()
+    
+        extended_gaussian_heatmap = extended_gaussian_heatmap + wave_gaussian_heatmap
+        
+        
+    if padding_target is not None:
+        left, right, top, bottom = padding_target
+        assert left<=w and right<=w and top<=h and bottom<=h,'Maximum padding allowed is (W,W,H,H)'
+        extended_gaussian_heatmap = extended_gaussian_heatmap[h-top:2*h+bottom, w-left:2*w+right]
+    
+    if normalize: extended_gaussian_heatmap = extended_gaussian_heatmap/extended_gaussian_heatmap.sum()
+    
+    filled_values = (extended_gaussian_heatmap>0).to(device)
+    
+    if make_plots:
+        fig, ax = plt.subplots(1, 3, figsize=(15, 15))
+        ax[0].imshow(generated_heatmap.detach().cpu())
+        ax[0].set_title('Original heatmap')
+        ax[1].imshow(filled_values.detach().cpu())
+        ax[1].set_title('Filled pixels')
+        ax[2].imshow(extended_gaussian_heatmap.detach().cpu())
+        ax[2].set_title('Final extended heatmap')
+        plt.show()
+    
+    if return_filled: return extended_gaussian_heatmap, filled_values
+    return extended_gaussian_heatmap
+
+def batched_gaussian_symmetric_wave_image(heatmaps, padding_target=None, dilation_iters=0, alpha=0.25, overwrite_original_pixels=False, normalize=True):
+    """Perform fast Gaussian completion on a batch of heatmaps based on the symmetric lines that cross the center of the Gaussian distribution (maximum pixel). This implementation makes use of coordinates masks for fast completion with an optional dilation completion process.
+
+    Args:
+        heatmaps (torch.Tensor): Heatmaps generated by the model with shape [B, E, H, W]
+        padding_target (tuple, default = None): Padding to be added and filled to the input heatmap in the form of the tuple (left, right, top, bottom). If None, then the tuple defaults to (W, W, H, H). Maximum padding is (W, W, H, H).
+        dilation_iters (int, default = 0): Number of times to repeat the process of heatmap dilation used to further extend the symmetrized heatmap.
+        alpha (float, default = 0.25): Additional value to be added to the denumerator during the averaging of the pixels within the pixel dilation process.
+        overwrite_original_pixels (bool, default = False): Whether to overwrite the pixels of the input heatmap. This can affect ellipse perfomance due to modifying the original heatmap.
+        normalize (bool, default = True): Whether to normalize the extended heatmap to be returned.
+
+    Returns:
+        extended_gaussian_heatmap (torch.Tensor): Extended heatmap containing a completed Gaussian distribution with shape [top + H + bottom, left + W + right]
+        filled_values (torch.Tensor): Optional boolean mask with all the pixels that have been filled over the extended heatmap with shape [top + H + bottom, left + W + right]
+    """
+    
+    batches, locations = heatmaps.size()[0:2]
+    all_heatmaps = []
+    
+    for batch_num in range(batches):
+        batch_heatmaps = []
+        
+        for location_num in range(locations):
+            extended_heatmap = gaussian_symmetric_wave_image(heatmaps[batch_num, location_num], padding_target=padding_target, dilation_iters=dilation_iters, alpha=alpha, overwrite_original_pixels=overwrite_original_pixels, normalize=normalize)
+            
+            batch_heatmaps.append( extended_heatmap )
+        
+        batch_heatmaps =  torch.stack(batch_heatmaps, dim=0)
+        all_heatmaps.append( batch_heatmaps )
+    
+    return torch.stack(all_heatmaps, dim=0)
+
